#define AXDEBUG 1
/*
 * Definitions for Amateur Radio packet protocol version 2
 * Per ARRL document
 */
#define	AX_VERSION	2

#define AX_MINHDR	15	/* 2 addrs + ctl */
#define AX_MAXHDR	72	/* 10 addrs + pid + ctl */
#define AX_MAXDATA	256	/* yeah, sure */
#define	AX_MAXPACKET	(AX_MAXHDR+AX_MAXDATA)	/* maximum packet size */

/*
 * Structure holding an AX.25 packet
 *
 * There is always a destination and source address.  The 8 possible
 * repeater fields are usually omitted, but we have to have room for them.
 *
 * In a data packet (type I or UI), following the control field
 * are a pid and from zero to 256 bytes of data
 *
 * we pass this header in one mbuf, the data chained after it
 *
 */

struct ax_addr {
	u_char	callsign[6];		/* left, space fill, uppercase */
	u_char	ssid;			/* 0-15 in 4 lowest bits */
};

/* note that this structure fits inside an mbuf
   and that the 10 addresses are contiguous at the front
   don't break that */
struct ax_hdr {
	struct ax_addr	ax_dst;		/* dest address */
	struct ax_addr	ax_src;		/* source address */
	struct ax_addr	ax_digi[8];	/* repeater addresses */
	u_char	ax_ctl;			/* control byte */
	u_char	ax_pid;			/* protocol, AX_PID */
	u_char	ax_rptcnt;		/* how many repeaters */
	int	ax_hlen;		/* length of header in packet */
	int	ax_dlen;		/* length of data in packet */
	u_char	ax_cmd;			/* true if it's a command frame */
	u_char	ax_rbits;		/* the reserved bits */
};

struct ax {				/* full packet is header and data */
	struct ax_hdr *hdr;
	u_char *data;
};

struct ax_arphdr {			/* header of an AX.25 arp packet */
	struct	ax_addr ax_dst;		/* will be QST-0 */
	struct	ax_addr	ax_src;		/* our address */
	u_char	ax_ctl;			/* control byte = UI */
	u_char	ax_pid;			/* protocol, AX_PID = ARP */
};

/*
 * Protocols
 */
#define	AXPROTO_AX		0		/* dummy for AX.25 */
#define	AXPROTO_DGRAM		1		/* UI */
#define	AXPROTO_STREAM		2		/* connected mode */
#define	AXPROTO_RAW		3		/* raw AX packet */
#define	AXPROTO_MAX		3

/*
 * Socket address, AX.25 style.
 */
struct sockaddr_ax {
	u_char	sax_len;
	u_char	sax_family;
	struct	ax_addr sax_addr;
	char	sax_zero[7];
};


/*
 * Options for use with [gs]etsockopt at the AX level.
 * First word of comment is data type; bool is stored in int.
 */
#define	AX_OPTIONS	1	/* buf/ip_opts; set/get AX per-packet options */
#define	AX_HDRINCL	2	/* int; header is included with data (raw) */
#define	AX_RETOPTS	8	/* ip_opts; set/get AX per-packet options */

/*
 * Interface address, AX25 version.  One of these structures
 * is allocated for each interface with an AX25 address.
 * The ifaddr structure contains the protocol-independent part
 * of the structure and is assumed to be first.
 */
struct ax_ifaddr {
	struct	ifaddr	ia_ifa;		/* protocol-independent info */
#define	ia_ifp		ia_ifa.ifa_ifp
#define ia_flags	ia_ifa.ifa_flags
	struct	ax_ifaddr *ia_next;	/* next in list of ax25 addresses */
	struct	sockaddr_ax ia_addr;	/* reserve space for interface name */
	struct	sockaddr_ax ia_broadaddr; /* reserve space for broadcast addr */
};

struct	ax_aliasreq {
	char	ifra_name[IFNAMSIZ];		/* if name, e.g. "en0" */
	struct	sockaddr_ax ifra_addr;
	struct	sockaddr_ax ifra_broadaddr;
	struct	sockaddr_ax ifra_mask;
};

/*
 * Given a pointer to an ax_ifaddr (ifaddr),
 * return a pointer to the addr as a sockaddr_ax.
 */
#define	IA_SAX(ia) (&(((struct ax_ifaddr *)(ia))->ia_addr))

struct	axstat {
	long	axs_total;		/* total packets received */
	long	axs_badsum;		/* checksum bad */
	long	axs_tooshort;		/* packet too short */
	long	axs_toosmall;		/* not enough data */
	long	axs_badhlen;		/* ax header length < data size */
	long	axs_badlen;		/* ax length < ax header length */
	long	axs_fragments;		/* fragments received */
	long	axs_fragdropped;	/* frags dropped (dups, out of space) */
	long	axs_fragtimeout;	/* fragments timed out */
	long	axs_forward;		/* packets forwarded */
	long	axs_cantforward;	/* packets rcvd for unreachable dest */
	long	axs_redirectsent;	/* packets forwarded on same net */
	long	axs_noproto;		/* unknown or unsupported protocol */
	long	axs_delivered;		/* packets consumed here */
	long	axs_localout;		/* total axpackets generated here */
	long	axs_odropped;		/* lost packets due to nobufs, etc. */
	long	axs_reassembled;	/* total packets reassembled ok */
	long	axs_fragmented;		/* output packets fragmented ok */
	long	axs_ofragments;		/* output fragments created */
	long	axs_cantfrag;		/* don't fragment flag was set, etc. */
};


/*
 * AX25 routing table.  Stored in mbufs; grows
 */
struct ax_rt {
	struct ax_addr	ax_dst;		/* dest address */
	struct ax_addr	ax_digi[8];	/* repeater addresses */
	u_char		ax_rptcnt;	/* how many repeaters */
	short		rt_ttl;		/* time to live -1=forever */
	struct ax_rt	*rt_next;	/* next entry pointer */
	struct ifnet	*rt_ifp;	/* the interface to use */
};

/*
 * Known values for the Protocol ID byte in ax.25 packets
 */
#define AXPID_TEXNET		0xc3	/* TexNet */
#define AXPID_LQ		0xc4	/* Link Quality */
#define AXPID_ATALK		0xca	/* AppleTalk */
#define AXPID_AARP		0xcb	/* AppleTalk ARP */
#define AXPID_IP		0xcc	/* Internet Protocol */
#define AXPID_ARP		0xcd	/* ARP */
#define AXPID_NETROM		0xcf	/* Net/Rom */
#define AXPID_NONE		0xf0	/* none (just a char stream) */

/*
 * values for the control byte
 * control byte maps to packet type after appropriate masking
 * you also get P/F bit, and window sequence numbers from it
 */

#define AXC_PF		0x10
#define AXC_I		0x01
#define AXC_S		0x0f
#define AXC_U		0xef
#define ax_pf(c)	(((c) & AXC_PF) >> 4)
#define	axt_i(c)	(((c) & AXC_I) == 0)
#define	axt_rr(c)	(((c) & AXC_S) == 1)
#define	axt_rnr(c)	(((c) & AXC_S) == 0x05)
#define	axt_rej(c)	(((c) & AXC_S) == 0x09)
#define	axt_sabm(c)	(((c) & AXC_U) == 0x2f)
#define	axt_disc(c)	(((c) & AXC_U) == 0x43)
#define	axt_dm(c)	(((c) & AXC_U) == 0x0f)
#define	axt_ua(c)	(((c) & AXC_U) == 0x63)
#define	axt_frmr(c)	(((c) & AXC_U) == 0x87)
#define	axt_ui(c)	(((c) & AXC_U) == 0x03)
#define AXC_NR		0xe0
#define AXC_NS		0x0e
#define	axp_nr(c)	(((c) & AXC_NR) >> 5)
#define	axp_ns(c)	(((c) & AXC_NS) >> 1)
/*
 * packet type manifest constants for the state machine
 */
#define	AXPT_I		0
#define	AXPT_RR		1
#define	AXPT_RNR	2
#define	AXPT_REJ	3
#define	AXPT_SABM	4
#define	AXPT_DISC	5
#define	AXPT_DM		6
#define	AXPT_UA		7
#define	AXPT_FRMR	8
#define	AXPT_UI		9
/* these are EXACTLY as above + 10 */
#define	AXPT_IPF	10
#define	AXPT_RRPF	11
#define	AXPT_RNRPF	12
#define	AXPT_REJPF	13
#define	AXPT_SABMPF	14
#define	AXPT_DISCPF	15
#define	AXPT_DMPF	16
#define	AXPT_UAPF	17
#define	AXPT_FRMRPF	18
#define	AXPT_UIPF	19
/* sent as commands */
#define	AXPT_RRc	21
#define	AXPT_RNRc	22

/*
 * event occurances - manifest constants
 */
#define AXE_BADFM	1
#define AXE_BADNR	2
#define AXE_BADNS	3
#define AXE_BUSY	4
#define AXE_CONN	5
#define AXE_DISC	6
#define AXE_N2		7
#define AXE_RDY		8
#define AXE_T1		9
#define AXE_T3		10

/*
 * packet response types manifest constants
 */
#define AXR_DISC	1
#define AXR_DM		2
#define AXR_FRMR	3
#define AXR_I		4
#define AXR_REJ		5
#define AXR_RNR		6
#define AXR_RNRc	7
#define AXR_RR		8
#define AXR_RRc		9
#define AXR_SABM	10
#define AXR_UA		11
#define AXR_UI		12
