#include "param.h"
#include "systm.h"
#include "malloc.h"
#include "mbuf.h"
#include "domain.h"
#include "protosw.h"
#include "socket.h"
#include "errno.h"
#include "time.h"
#include "kernel.h"
#include "../net/if.h"

#include "ax.h"
#include "ax_var.h"
#include "ax_debug.h"

extern int axptype();

struct ax_ccb *axfindccb();
struct ax_ccb *axnewccb();

/*
 * Implements the AX.25 version 2.0 protocol state machine as given
 * in appendix D of the ARRL protocol document, more or less
 */
axfsm(ah, ccb)
struct ax_hdr *ah;
struct ax_ccb *ccb;
{
	int newstate;
	int ptype, mode;
	int ns, nr, i;

	if (ccb == 0 || ccb == &axb)
		return;

 	nr = axp_nr(ah->ax_ctl);
 	ns = axp_ns(ah->ax_ctl);
	ptype = axpacketype(ah->ax_ctl);
	if (ptype == AXPT_UI || ptype == AXPT_UIPF)
		mode = AXPROTO_DGRAM;
	else
		mode = AXPROTO_STREAM;

	ccb->rxhdr = ah;

#ifdef AXDEBUG
	printf("axFSM: ctl=%02x ptype=%d mode=%d cmd=%d ns=%d nr=%d\n",
		 ah->ax_ctl, ptype, mode, ah->ax_cmd, ns, nr);
	axaddrprint("src", ah->ax_src);
	axaddrprint("dst", ah->ax_dst);
	axaddrprint(" ccb mycall", ccb->mycall);
	axaddrprint(" ccb hiscall", ccb->hiscall);
	printf(" ax_ccbnext; %x\n", ccb->ax_ccbnext);
	printf(" state %d", ccb->state);
	if (ccb->state >= 0 && ccb->state <= 16)
		printf(" (%s)\n", axstates[ccb->state]);
	else
		printf(" (unknown)\n");
	printf(" vr %d\n", ccb->vr);
	printf(" vs %d\n", ccb->vs);
	printf(" t1 %d\n", ccb->t1);
	printf(" t2 %d\n", ccb->t2);
	printf(" t3 %d\n", ccb->t3);
	printf(" n2 %d\n", ccb->n2);
	printf(" k %d\n", ccb->k);
	printf(" paclen %d\n", ccb->paclen);
	printf(" retries %d\n", ccb->retries);
	printf(" maxframe %d\n", ccb->maxframe);
	printf(" keepalive %d\n", ccb->keepalive);
#endif

	/* keep a copy of ns,nr in the ccb - window control */
	ccb->ns = ns;
	ccb->nr = nr;

	/* any valid packet will clear the t1 timer */
	ccb->t1 = 0;

	/* frame in seq? 2.4.4.3 */
	if (ptype == AXPT_I) {
		if (ccb->vr == ns) {
			ccb->vr = (ccb->vr + 1) % 8;
		} else {
			axevent(AXE_BADNS, ccb);
			return(1);
		}
	}

	/* does this frame ACK any outstanding I frames? */
	for (i=nr; i < nr+7; i++)	/* max windowsize is 7 */
		if (ccb->txp[i%8]) {	/* yup, was an outstanding frame */
			m_freem(ccb->txp[i%8]);
			ccb->txp[i%8] = 0;
			ccb->k--;
		}
	for (i=nr; i < nr+8; i++)
		if (ccb->txp[i%8]) {	/* still some outstanding frames */
/* XXX resend an unack'd frame and restart timer */
			break;
		}

	if (ptype == AXPT_REJ || ptype == AXPT_REJPF)
		ccb->vs = nr;

	newstate = 0;

	if (ah->ax_cmd)
		{
		switch(ccb->state) {
		default:
			printf("Unknown state %d\n", ccb->state);
			break;
		case 1:
			switch(ptype) {
			case AXPT_IPF:
			case AXPT_RRPF:
			case AXPT_REJPF:
			case AXPT_RNRPF:
				ax_response(ccb, AXR_DM);
				break;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 5;
				break;
			case AXPT_DISCPF:
				ax_response(ccb, AXR_DM);
				break;
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				break;
			} /* ptype switch */
			break;
		case 2:
			switch(ptype) {
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 5;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_DM);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 3:
			switch(ptype) {
			case AXPT_IPF:
			case AXPT_RRPF:
			case AXPT_RNRPF:
			case AXPT_REJPF:
				ax_response(ccb, AXR_FRMR);
				break;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 5;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_DM);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 4:
			switch(ptype) {
			case AXPT_IPF:
			case AXPT_RRPF:
			case AXPT_RNRPF:
			case AXPT_REJPF	:
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_DM);
				newstate = 1;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 5:
			switch(ptype) {
			case AXPT_IPF:
			case AXPT_RRPF:
			case AXPT_REJPF	:
				ax_response(ccb, AXR_RR);
				break;
			case AXPT_I:
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				break;
			case AXPT_RNRPF:
				ax_response(ccb, AXR_RR);
			case AXPT_RNR:
				newstate = 9;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				newstate = 1;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				ccb->vs = ccb->vr = 0;
				break;
			} /* ptype switch */
			break;
		case 6:
			switch(ptype) {
			case AXPT_IPF:
				ax_response(ccb, AXR_RR);
				newstate = 5;
				break;
			case AXPT_I:
				ax_response(ccb, AXR_I);
				newstate = 5;
				break;
			case AXPT_RNRPF:
				newstate = 15;
			case AXPT_RRPF:
			case AXPT_REJPF	:
				ax_response(ccb, AXR_RR);
				break;
			case AXPT_RNR:
				newstate = 15;
			case AXPT_REJ:
			case AXPT_RR:
				ax_response(ccb, AXR_I);
				break;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 5;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 7:
			switch(ptype) {
			case AXPT_RNRPF:
				newstate = 12;
			case AXPT_IPF:
			case AXPT_REJPF:
			case AXPT_RRPF:
				ax_response(ccb, AXR_RR);
				break;
			case AXPT_RNR:
				newstate = 12;
			case AXPT_RR:
			case AXPT_I:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				break;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 5;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 8:
			switch(ptype) {
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				break;
			case AXPT_RNRPF:
				ax_response(ccb, AXR_RNR);
			case AXPT_RNR:
				newstate = 10;
				break;
			case AXPT_IPF:
			case AXPT_I:
			case AXPT_RRPF:
			case AXPT_REJPF	:
				ax_response(ccb, AXR_RNR);
				break;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 9:
			switch(ptype) {
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				newstate = 5;
				break;
			case AXPT_RRPF:
			case AXPT_REJPF	:
				newstate = 5;
			case AXPT_RNRPF:
			case AXPT_IPF:
			case AXPT_I:
				ax_response(ccb, AXR_RR);
				break;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 5;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 10:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_REJPF	:
				newstate = 8;
			case AXPT_IPF:
			case AXPT_I:
			case AXPT_RNRPF:
				ax_response(ccb, AXR_RNR);
				break;
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				newstate = 8;
				break;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 8;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 11:
			switch(ptype) {
			case AXPT_IPF:
			case AXPT_I:
			case AXPT_RRPF:
			case AXPT_RR:
			case AXPT_RNRPF:
			case AXPT_RNR:
			case AXPT_REJPF	:
			case AXPT_REJ:
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 8;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 12:
			switch(ptype) {
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				newstate = 7;
				break;
			case AXPT_RRPF:
			case AXPT_REJPF	:
				newstate = 7;
			case AXPT_IPF:
			case AXPT_I:
			case AXPT_RNRPF:
				ax_response(ccb, AXR_RR);
				break;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 5;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 13:
			switch(ptype) {
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				newstate = 11;
				break;
			case AXPT_RRPF:
			case AXPT_REJPF	:
				newstate = 11;
			case AXPT_IPF:
			case AXPT_I:
			case AXPT_RNRPF:
				ax_response(ccb, AXR_RNR);
				break;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 8;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 14:
			switch(ptype) {
			case AXPT_RNR:
				newstate = 16;
				break;
			case AXPT_RNRPF:
				newstate = 16;
			case AXPT_IPF:
			case AXPT_I:
			case AXPT_RRPF:
			case AXPT_REJPF	:
				ax_response(ccb, AXR_RNR);
				break;
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				break;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 8;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 15:
			switch(ptype) {
			case AXPT_IPF:
			case AXPT_I:
				ax_response(ccb, AXR_I);
				newstate = 9;
				break;
			case AXPT_RRPF:
			case AXPT_REJPF:
				newstate = 6;
			case AXPT_RNRPF:
				ax_response(ccb, AXR_RR);
				break;
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				break;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 5;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 16:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_REJPF	:
				newstate = 14;
			case AXPT_IPF:
			case AXPT_I:
			case AXPT_RNRPF:
				ax_response(ccb, AXR_RNR);
				break;
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				newstate = 14;
				break;
			case AXPT_SABMPF:
			case AXPT_SABM:
				ax_response(ccb, AXR_UA);
				newstate = 5;
				break;
			case AXPT_DISCPF:
			case AXPT_DISC:
				ax_response(ccb, AXR_UA);
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		} /* ccb->state switch */

	} else {	/* NOT a command frame */
		switch(ccb->state) {
		default:
			printf("Unknown ccb->state %d\n", ccb->state);
			break;
		case 1:
			break;
		case 2:
			switch(ptype) {
			case AXPT_UA:
			case AXPT_UAPF:
				newstate = 5;
				ccb->t1 = 0;	/* cancel timer */
				break;
			case AXPT_DM:
			case AXPT_DMPF:
				newstate = 1;
				break;
			} /* ptype switch */
			break;
		case 3:
			switch(ptype) {
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		case 4:
			switch(ptype) {
			case AXPT_UA:
			case AXPT_UAPF:
			case AXPT_DM:
			case AXPT_DMPF:
				newstate = 1;
			} /* ptype switch */
			break;
		case 5:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_REJPF:
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				break;
			case AXPT_RNRPF:
			case AXPT_RNR:
				newstate = 9;
				break;
			case AXPT_DM:
			case AXPT_DMPF:
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		case 6:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_RR:
			case AXPT_REJPF:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				break;
			case AXPT_RNRPF:
			case AXPT_RNR:
				newstate = 15;
				break;
			case AXPT_UA:
			case AXPT_UAPF:
			case AXPT_DM:
			case AXPT_DMPF:
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		case 7:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_REJPF:
				newstate = 5;
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				break;
			case AXPT_RNRPF:
				newstate = 9;
				break;
			case AXPT_RNR:
				newstate = 12;
				break;
			case AXPT_UA:
			case AXPT_UAPF:
			case AXPT_DM:
			case AXPT_DMPF:
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		case 8:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_RR:
			case AXPT_REJPF:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				break;
			case AXPT_RNRPF:
			case AXPT_RNR:
				newstate = 10;
				break;
			case AXPT_UA:
			case AXPT_UAPF:
			case AXPT_DM:
			case AXPT_DMPF:
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		case 9:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_RR:
			case AXPT_REJPF:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				newstate = 5;
				break;
			case AXPT_UA:
			case AXPT_UAPF:
			case AXPT_DM:
			case AXPT_DMPF:
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		case 10:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_RR:
			case AXPT_REJPF:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				newstate = 8;
				break;
			case AXPT_UA:
			case AXPT_UAPF:
			case AXPT_DM:
			case AXPT_DMPF:
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		case 11:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_REJPF:
				newstate = 8;
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				break;
			case AXPT_RNRPF:
				newstate = 10;
				break;
			case AXPT_RNR:
				newstate = 13;
				break;
			case AXPT_UA:
			case AXPT_UAPF:
			case AXPT_DM:
			case AXPT_DMPF:
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		case 12:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_REJPF:
				ax_response(ccb, AXR_I);
				newstate = 5;
				break;
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				newstate = 7;
				break;
			case AXPT_UA:
			case AXPT_UAPF:
			case AXPT_DM:
			case AXPT_DMPF:
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		case 13:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_REJPF:
				ax_response(ccb, AXR_I);
				newstate = 8;
				break;
			case AXPT_RR:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				newstate = 11;
				break;
			case AXPT_UA:
			case AXPT_UAPF:
			case AXPT_DM:
			case AXPT_DMPF:
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		case 14:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_RR:
			case AXPT_REJPF:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				break;
			case AXPT_RNRPF:
			case AXPT_RNR:
				newstate = 16;
				break;
			case AXPT_UA:
			case AXPT_UAPF:
			case AXPT_DM:
			case AXPT_DMPF:
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		case 15:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_RR:
			case AXPT_REJPF:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				newstate = 6;
				break;
			case AXPT_UA:
			case AXPT_UAPF:
			case AXPT_DM:
			case AXPT_DMPF:
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		case 16:
			switch(ptype) {
			case AXPT_RRPF:
			case AXPT_RR:
			case AXPT_REJPF:
			case AXPT_REJ:
				ax_response(ccb, AXR_I);
				newstate = 6;
				break;
			case AXPT_UA:
			case AXPT_UAPF:
			case AXPT_DM:
			case AXPT_DMPF:
			case AXPT_FRMR:
			case AXPT_FRMRPF:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ptype switch */
			break;
		} /* state switch */
	} /* end response */

	if (newstate) {
#ifdef AXDEBUG
		printf("state %d -> %d", ccb->state, newstate);
		if (newstate >= 0 && newstate <= 16)
			printf(" (%s)\n", axstates[newstate]);
		else
			printf(" (unknown)\n");
#endif
		axupdsockstat(ccb,newstate);
		ccb->state = newstate;
#ifdef maybenot
		if (newstate == 1)
			axdelccb(ccb);
#endif
	}
	return(0);
}

/*
 * state machine changes in response to events, not frames
 */
axevent(event,ccb)
int event;
struct ax_ccb *ccb;
{
	int newstate = 0;

#ifdef AXDEBUG
	printf("axevent %d (%s) state=%d", event, axevents[event], ccb->state);
	if (ccb->state >= 0 && ccb->state <= 16)
		printf(" (%s)\n", axstates[ccb->state]);
	else
		printf(" (unknown)\n");
#endif

	if (event) {
		switch(event) {
		case AXE_CONN:
			if (ccb->state != 2) {
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			}
			break;
		case AXE_DISC:
			if (ccb->state != 1 && ccb->state != 4) {
				ax_response(ccb, AXR_DISC);
				newstate = 4;
			}
			break;
		case AXE_BUSY:
			switch(ccb->state) {
			case 5:
				ax_response(ccb, AXR_RNR);
				newstate = 8;
				break;
			case 6:
				ax_response(ccb, AXR_RNR);
				newstate = 14;
				break;
			case 7:
				ax_response(ccb, AXR_RNR);
				newstate = 11;
				break;
			case 9:
				ax_response(ccb, AXR_RNR);
				newstate = 10;
				break;
			case 12:
				ax_response(ccb, AXR_RNR);
				newstate = 13;
				break;
			case 15:
				ax_response(ccb, AXR_RNR);
				newstate = 16;
				break;
			} /* ccb->state switch */
			break;
		case AXE_RDY:
			switch(ccb->state) {
			case 8:
				ax_response(ccb, AXR_RR);
				newstate = 5;
				break;
			case 10:
				ax_response(ccb, AXR_RR);
				newstate = 9;
				break;
			case 11:
				ax_response(ccb, AXR_RR);
				newstate = 7;
				break;
			case 13:
				ax_response(ccb, AXR_RR);
				newstate = 12;
				break;
			case 14:
				ax_response(ccb, AXR_RR);
				newstate = 6;
				break;
			case 16:
				ax_response(ccb, AXR_RR);
				newstate = 15;
				break;
			} /* ccb->state switch */
			break;
		case AXE_T1:
			if (++ccb->n2 > ccb->retries) {
				axevent(AXE_N2, ccb);
				break;
			}

			switch(ccb->state) {
			case 2:
				ax_response(ccb, AXR_SABM);
				break;
			case 3:
				ax_response(ccb, AXR_FRMR);
				break;
			case 4:
				ax_response(ccb, AXR_DISC);
				break;
			case 5:
			case 6:
				newstate = 7;
			case 7:
			case 12:
				ax_response(ccb, AXR_RRc);
				break;
			case 8:
			case 14:
			case 16:
				ax_response(ccb, AXR_RNRc);
				newstate = 11;
				break;
			case 9:
			case 15:
				ax_response(ccb, AXR_RRc);
				newstate = 12;
				break;
			case 10:
				newstate = 13;
			case 11:
			case 13:
				ax_response(ccb, AXR_RNRc);
				break;
			} /* ccb->state switch */
			break;
		case AXE_T3:
			switch(ccb->state) {
			case 2:
				ax_response(ccb, AXR_SABM);
				break;
			case 3:
				ax_response(ccb, AXR_FRMR);
				break;
			case 4:
				ax_response(ccb, AXR_DISC);
				break;
			case 5:
			case 6:
				ax_response(ccb, AXR_RRc);
				newstate = 7;
				break;
			case 8:
			case 14:
			case 16:
				ax_response(ccb, AXR_RNRc);
				newstate = 11;
				break;
			case 9:
			case 15:
				ax_response(ccb, AXR_RRc);
				newstate = 12;
				break;
			case 10:
				ax_response(ccb, AXR_RNRc);
				newstate = 13;
				break;
			} /* ccb->state switch */
			break;
		case AXE_N2:
			switch(ccb->state) {
			case 2:
			case 4:
				newstate = 1;
				break;
			case 3:
			case 6:
			case 7:
			case 11:
			case 12:
			case 13:
				ax_response(ccb, AXR_SABM);
				newstate = 2;
			} /* ccb->state switch */
			break;
		case AXE_BADNS:
			switch(ccb->state) {
			case 5:
				ax_response(ccb, AXR_REJ);
				newstate = 6;
				break;
			case 9:
				ax_response(ccb, AXR_REJ);
				newstate = 15;
				break;
			case 8:
			case 10:
			case 14:
				ax_response(ccb, AXR_RNR);
			} /* ccb->state switch */
			break;
		case AXE_BADNR:
		case AXE_BADFM:
			if (ccb->state > 4) {
				ax_response(ccb, AXR_FRMR);
				newstate = 3;
			}
			break;
		} /* event switch */
	} /* event */

	if (newstate) {
#ifdef AXDEBUG
		printf("eventstate %d -> %d", ccb->state, newstate);
		if (newstate >= 0 && newstate <= 16)
			printf(" (%s)\n", axstates[newstate]);
		else
			printf(" (unknown)\n");
#endif
		axupdsockstat(ccb,newstate);
		ccb->state = newstate;
		if (newstate == 1)
			axdelccb(ccb);
	}
	return(0);
}


/* update the socket status */
axupdsockstat(ccb,newstate)
struct ax_ccb *ccb;
int newstate;
{
	if (ccb->ax_socket)
		switch(newstate) {
		case 1:				/* became disconnected */
			soisdisconnected(ccb->ax_socket);
			printf("soisdisconnected\n");
			break;
		case 4:				/* disconnecting */
			soisdisconnecting(ccb->ax_socket);
			printf("soisdisconnecting\n");
			break;
		case 2:				/* connecting */
			soisconnecting(ccb->ax_socket);
			printf("soisconnecting\n");
			break;
		case 5:				/* connected */
			if (ccb->state == 2) {
				soisconnected(ccb->ax_socket);
				printf("soisconnected\n");
			}
			break;
		}
}
