#include "param.h"
#include "malloc.h"
#include "mbuf.h"
#include "errno.h"
#include "protosw.h"
#include "socket.h"
#include "socketvar.h"

#include "../net/if.h"

#include "ax.h"

#include "if_ki.h"

struct ax_rt ax_route();

extern char *axresponses[];
extern struct ki_softc ki_softc[];

/*
 * AX output.
 * am contains the parsed-out header that will form the packet header
 * m contains data field for UI and I frames, info for FRMR
 * m is passed or freed
 * am is freed
 */
ax_output(am, m, flags)
	struct mbuf *am;		/* header struct of packet */
	struct mbuf *m;			/* data field of packet if any */
	int flags;
{
	u_char *dp, *ap;
	struct ax_ifaddr *aa;
	struct ax_hdr *ah;
	struct ifnet *ifp;
	struct ax_rt *rt;
	struct sockaddr_ax dst;
	int alen, hlen, i;
	int error = 0;
	/*
	 * construct the packet, header first
	 */
	ah = mtod(am, struct ax_hdr *);
 	/* address size is src+dst+digipeaters */
 	alen = (2 + ah->ax_rptcnt) * sizeof(struct ax_addr);
 	/* header is one longer than addresses for ctl byte */
 	hlen = alen + 1;
 	/* plus pid byte if it's an I or UI frame */
 	if (axt_i(ah->ax_ctl) || axt_ui(ah->ax_ctl)) {
 		hlen++;
	}

	if (m == 0) {			/* no data */
 		MGET(m, M_WAIT, MT_DATA);	/* header fits in an mbuf */
		m->m_len = hlen;
	} else {
 		M_PREPEND(m, hlen, M_WAIT);
	}
 	if (m == 0)		/* no mem to prepend header */
 		return(ENOBUFS);

	dp = mtod(m, u_char *);	/* start of header area */
	ap = (u_char *)ah;	/* addresses in header struct */
	/* copy over addresses, left-shifting as you go */
	for (i = 0; i < alen; i++)
		*dp++ = (*ap++ << 1);
 	/* set end-of-address flag */
 	*(dp-1) |= 1;
 	/* plus control byte */
 	*dp++ = ah->ax_ctl;
 	/* plus PID if it's an I or UI frame */
 	if (axt_i(ah->ax_ctl) || axt_ui(ah->ax_ctl))
 		*dp++ = ah->ax_pid;

#ifdef AXDEBUG
	printf("ax_output: packet with %d bytes of header:\n", hlen);
	dp = mtod(m, u_char *);	/* start of header area */
	for (i=0; i<hlen; i++)
		printf("%02x ", *dp++);
	printf("\n");
#endif

	dst.sax_len = sizeof(struct sockaddr_ax);
	dst.sax_family = AF_AX25;
	dst.sax_addr = ah->ax_dst;

	/*
	 * somehow figure out which AX.25 interface it's to go out
	 */
#ifdef notyet
	rt = ax_route(ah);
	ifp = rt->rt_ifp;
#endif /* notyet */
	ifp = &ki_softc[0].kc_if;

	/* ship it to the interface */
#ifdef AXDEBUG
	printf("ax_output: shipping to %s on interface %s%d\n",
		axaddrstr(dst.sax_addr), ifp->if_name, ifp->if_unit);
#endif
	error = (*ifp->if_output)(ifp, m, &dst);
	/* we're all done with the header struct, so free it */
	ah = 0;
	m_free(am);
	return (error);
}

/*
 * make up a packet and pass to output
 */ 
ax_send(dst, src, pt, pid, pf, ns, nr, cmd, m, flags)
struct ax_addr dst, src;
u_char pt, pid, pf, ns, nr, cmd;
struct mbuf *m;
int flags;
{
	struct mbuf *am;
	struct ax_hdr *ah;

#ifdef AXDEBUG
	printf("ax_send pt=%d (%s) pf=%d pid=%02x ns=%d nr=%d cmd=%d flags=%d\n",
		pt, axresponses[pt], pf, pid, ns, nr, cmd, flags);
	axaddrprint("   dst", dst);
	axaddrprint("   src", src);
	if (m)
		printf("   data length %d\n", m->m_len);
#endif

 	MGET(am, M_WAIT, MT_DATA);	/* header struct fits in an mbuf */
 	if (am == 0) {
 		printf("ax_send no mbufs\n");
 		m_freem(m);
		return(ENOBUFS);
	}
 	M_ALIGN(am, sizeof(struct ax_hdr));
	am->m_len = sizeof(struct ax_hdr);
 	ah = mtod(am, struct ax_hdr *);
 	bzero(ah, sizeof(struct ax_hdr));

 	/* fill up the header struct with the right data */
 	ah->ax_dst = dst;
	ah->ax_dst.ssid |= 0x30;	/* set reserved bits on */
 	ah->ax_src = src;
	ah->ax_src.ssid |= 0x30;	/* implied (not stated) by spec */
 	ah->ax_rptcnt = 0;		/* until routing tells us otherwise */

	ns = (ns & 7) << 1;
	pf = (pf & 1) << 4;
	nr = (nr & 7) << 5;

 	/* the control bytes */
 	switch (pt) {
case AXR_SABM:
		cmd = 1;
		ah->ax_ctl = 0x2f | pf;
		break;
case AXR_UA:
		cmd = 0;
		ah->ax_ctl = 0x63 | pf;
		break;
case AXR_DISC:
		cmd = 1;
		ah->ax_ctl = 0x43 | pf;
		break;
case AXR_DM:
		cmd = 0;
		ah->ax_ctl = 0x0f | pf;
		break;
case AXR_FRMR:
		cmd = 0;
		ah->ax_ctl = 0x87 | pf;
		break;
case AXR_REJ:
		cmd = 0;
		ah->ax_ctl = 9 | nr | pf;
		break;
case AXR_RNRc:
		cmd = 1;
		ah->ax_ctl = 5 | nr | pf;
		break;
case AXR_RNR:
		cmd = 0;
		ah->ax_ctl = 5 | nr | pf;
		break;
case AXR_RRc:
		cmd = 1;
		ah->ax_ctl = 1 | nr | pf;
		break;
case AXR_RR:
		cmd = 0;
		ah->ax_ctl = 1 | nr | pf;
		break;
case AXR_I:
		ah->ax_ctl = 0x01 | ns | nr | pf;
		break;
case AXR_UI:
		ah->ax_ctl = 0x03 | pf;
		break;
	} /* switch */

	ah->ax_pid = pid;

	if (cmd) {
		ah->ax_dst.ssid |= 0x40;	/* set command bit */
		ah->ax_src.ssid &= ~0x40;	/* unset response bit */
	} else {
		ah->ax_src.ssid |= 0x40;	/* set response bit */
		ah->ax_dst.ssid &= ~0x40;	/* unset command bit */
	}

	return(ax_output(am, m, flags));
}
